package net.kronos.mockup;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import net.kronos.mockup.exceptions.NoSuchHandlerException;
import net.kronos.mockup.handler.*;
import net.kronos.mockup.annotation.*;
import net.kronos.orm.rowset.ExceptionHandler;

/**
 * データを詰め込む
 * @author dai.yamamoto
 */
public class MockDataFactory {
	/** アノテーションとその処理ハンドラをセットで持つMap */
	protected Map<Class<? extends Annotation>,IMockHandlerCommand> annotationMap = new Hashtable<Class<? extends Annotation>,IMockHandlerCommand>();

	/**
	 * コンストラクタ
	 */
	public MockDataFactory(){
		this.addCustomAnnotation(MockPersonName.class, new PersonNameHandler());
		this.addCustomAnnotation(MockNumber.class, new NumberHandler());
		this.addCustomAnnotation(MockDatetime.class, new DateHandler());
		this.addCustomAnnotation(MockAddress.class, new AddressHandler());
		
	}
	
	/**
	 * アノテーションクラスを指定することで、ペアであるハンドラクラスを取得します。
	 * @param annClass
	 * @return
	 * @throws NoSuchHandlerException
	 */
	protected IMockHandlerCommand getAnnotationHandleCommand(Class<? extends Annotation> annClass) throws NoSuchHandlerException{
		if(this.annotationMap.containsKey(annClass) == false) 
			throw new NoSuchHandlerException();
		return this.annotationMap.get(annClass);
	}

	/**
	 * オブジェクトのフィールドにモックデータをセットする。
	 * @param <T>
	 * @param instance 返却するインスタンス
	 * @param fields アノテーションがセットされたフィールド
	 * @param count モックデータの数
	 * @return モックデータを詰め込んだオブジェクト
	 */
	private<T> T injectMockToField(T instance , Field[] fields ,int count){
		for(Field f : fields){
			Annotation[] anns =  f.getDeclaredAnnotations();
			for(Annotation an : anns){
				try{
					IMockHandlerCommand handler = getAnnotationHandleCommand(an.annotationType());
					boolean orgAccessible =f.isAccessible();
					
					// ハンドラの呼び出し
					Object value = handler.execute(an, f, instance,count);
					
					f.setAccessible(true);
					f.set(instance, value); // 値のセット
					f.setAccessible(orgAccessible);

					break;
				}catch(NoSuchHandlerException nshe ){
					continue;
				} catch (IllegalArgumentException e) {
					throw ExceptionHandler.wrapNoCheckException(e);
				} catch (IllegalAccessException e) {
					throw ExceptionHandler.wrapNoCheckException(e);
				}
			}
		}
		return instance;
	}
	
	/**
	 * モックデータを詰め込んだリストを返す。
	 * @param <T> リストに詰め込む型、clazz引数の型引数と一致
	 * @param clazz リストに詰め込む型
	 * @param objectList 元となるオブジェクトリスト
	 * @return モックデータを詰め込んだリスト
	 */
	public<T> List<T> inject(Class<T> clazz ,List<T> objectList ){
		return inject(clazz ,objectList ,objectList.size());
	}
	
	/**
	 * モックデータを詰め込んだリストを返す。
	 * @param <T> リストに詰め込む型、clazz引数の型引数と一致
	 * @param clazz リストに詰め込む型
	 * @param size 返すリストのサイズ
	 * @return モックデータを詰め込んだリスト
	 */
	public<T> List<T> inject(Class<T> clazz ,int size) {
		List<T> objectList = new ArrayList<T>();
		return inject(clazz,objectList,size);
	}

	/**
	 * モックデータを詰め込んだリストを返す。
	 * @param <T> リストに詰め込む型、clazz引数の型引数と一致
	 * @param clazz リストに詰め込む型
	 * @param size 返すリストのサイズ
	 * @param fieldList フィールドとアノテーションをマッピングしたリスト
	 * @return モックデータを詰め込んだリスト
	 */
	public<T> List<T> inject(Class<T> clazz ,int size,List<FieldAnnotationMapper> fieldList) {
		List<T> objectList = new ArrayList<T>();
		try {
			for(int i=0;i< size;i++){
				T instance = clazz.newInstance();
				
				List<Field> fieldsList =new ArrayList<Field>();
				for(FieldAnnotationMapper mapper : fieldList){
					try{
						Field f = clazz.getDeclaredField(mapper.getFieldName());
						fieldsList.add(f);
					} catch (NoSuchFieldException e) {
						// exception occur
					}
				}
				Field[] fields = {};
				fields = fieldsList.toArray(fields);

				instance = injectMockToField(instance,fields,i);
				objectList.add(instance);
			}
			return objectList;
		} catch (InstantiationException e) {
			throw ExceptionHandler.wrapNoCheckException(e);
		} catch (IllegalAccessException e) {
			throw ExceptionHandler.wrapNoCheckException(e);
		} catch (SecurityException e) {
			throw ExceptionHandler.wrapNoCheckException(e);
		}
	}	
	
	/**
	 * モックデータを詰め込んだリストを返す。
	 * @param <T> リストに詰め込む型、clazz引数の型引数と一致
	 * @param clazz リストに詰め込む型
	 * @param objectList 元となるオブジェクトリスト
	 * @param size 返すリストのサイズ
	 * @return モックデータを詰め込んだリスト
	 */
	public<T> List<T> inject(Class<T> clazz ,List<T> objectList ,int size){
		try {
			for(int i=0;i< size;i++){
				T instance = clazz.newInstance();
				Field[] fields = clazz.getDeclaredFields();

				instance = injectMockToField(instance,fields,i);
				objectList.add(instance);
			}
			return objectList;
		} catch (InstantiationException e) {
			throw ExceptionHandler.wrapNoCheckException(e);
		} catch (IllegalAccessException e) {
			throw ExceptionHandler.wrapNoCheckException(e);
		}
	}
	
	/**
	 * アノテーションとそれに対応するハンドラをセットする。
	 * @param ann モックデータアノテーション
	 * @param command コマンド
	 */
	protected void addCustomAnnotation(Class<? extends Annotation> ann ,IMockHandlerCommand command ){
		this.annotationMap.put(ann, command);
	}
	
}
